/**
 * @template T
 * @typedef {new (...args: any[]) => T} Newable<T>
 */

/**
 * @typedef {
    Int8Array
  | Uint8Array
  | Uint8ClampedArray
  | Int16Array
  | Uint16Array
  | Int32Array
  | Uint32Array
  | Float32Array
  | Float64Array
  | BigInt64Array
  | BigUint64Array
  } TypedArray
 */

/**
 * @typedef {
    Int8ArrayConstructor
  | Uint8ArrayConstructor
  | Uint8ClampedArrayConstructor
  | Int16ArrayConstructor
  | Uint16ArrayConstructor
  | Int32ArrayConstructor
  | Uint32ArrayConstructor
  | Float32ArrayConstructor
  | Float64ArrayConstructor
  | BigInt64ArrayConstructor
  | BigUint64ArrayConstructor
  } TypedArrayConstructor
 */

/**
 * @template {number} T
 * @typedef {`${T}` extends `-${string}` | '0' | `${string}.${string}` ? never : T } PositiveInteger
 */

/**
 * @template {ReadonlyArray<string>} T
 * @typedef {T extends readonly [] ? [] : T extends readonly [infer Head extends string, ... infer Tail extends ReadonlyArray<string>] ? [Capitalize<Head>, ...CapitalizeAll<Tail>] : string[]} CapitalizeAll
 */

/**
 * @template {ReadonlyArray<string>} T
 * @template {string} [Sep='']
 * @typedef {T extends readonly [] ? '' : T extends readonly [infer Head] ? Head : T extends readonly [infer Head extends string, ...infer Tail extends ReadonlyArray<string>] ? `${Head}${Sep}${Join<Tail, Sep>}` : string} Join
 */

/**
 * @template {string} S
 * @template {string} D
 * @typedef {D extends '' ? [S] : S extends `${infer Head}${D}${infer Tail}` ? [Head, ...Split<Tail, D>] : [S]} Split
 */

/*
 * @typedef {'-' | '_' | ' '} WordSeparator
 */

/**
 * @template {string} S
 * @template {string} D
 * @typedef {S extends `${infer Left}${D}${infer Right}` ? `${PascalCaseFromDelimiter<Left, D>}${Capitalize<PascalCaseFromDelimiter<Right, D>>}` : S} PascalCaseFromDelimiter
 */

/**
 * @template {string} S
 * @typedef {Capitalize<PascalCaseFromDelimiter<PascalCaseFromDelimiter<PascalCaseFromDelimiter<Lowercase<S>, '-'>, '_'>, ' '>>} PascalCaseString
 */

/**
 * @template {string | ReadonlyArray<string>} T
 * @typedef {T extends String ? PascalCaseString<T> : T extends ReadonlyArray<string> ? Join<CapitalizeAll<T>, ''> : T} PascalCase
 */

/**
 * @template {string} T
 * @param {T} s
 * @returns {Uppercase<T>}
 */
export const uppercase = s => /** @type {Uppercase<T>} */(s.toUpperCase())

/**
 * @template {string} T
 * @param {T} s
 * @returns {Lowercase<T>}
 */
export const lowercase = s => /** @type {Lowercase<T>} */(s.toLowerCase())

/**
 * @template {string} const T
 * @param {T} s
 * @returns {Split<T, '-'>}
 */
export const fromKebab = s => /** @type {Split<T, '-'>} */(s.split('-'))

/**
 * @template {string} const T
 * @param {T} xs
 * @returns {PascalCase<T>}
 */
export const toPascal = xs => /** @type {PascalCase<T>} */(uppercase(xs[0]) + xs.slice(1))

/**
 * @template {string[]} const T
 * @param {T} xs
 * @returns {PascalCase<T>}
 */
const pascal = xs => /** @type {PascalCase<T>} */(xs.map(toPascal).join(''))

/**
 * @template {string} const T
 * @param {...T} values
 * @returns {Readonly<{ [K in T as PascalCase<K>]: K }>}
 */
export const Enum = (...values) => /** @type {Readonly<{ [K in T as PascalCase<K>]: K }>} */(Object.freeze(
  values.reduce((acc, x) => {
    const key = pascal(fromKebab(x)).toString()
    acc[key] = x
    return acc
  }, {})
))


/** @typedef {(...args: any) => void} Listener */
export class EventEmitter {
  _listeners = {}

  /**
   * @param {PropertyKey} event 
   * @param {Listener} callback 
   */
  on(event, callback) {
    this._listeners[event] = this._listeners[event] || []
    this._listeners[event].push(callback)
  }

  /**
   * @param {PropertyKey} event 
   * @param {...any} args 
   */
  emit(event, ...args) {
    const listeners = this._listeners[event]

    if (listeners) {
      listeners.forEach(
        /** @param {Listener} cb */
        cb => cb(...args)
      )
    }
  }

  /**
   * @param {PropertyKey} event 
   * @param {Listener} callback 
   */
  off(event, callback) {
    const listeners = this._listeners[event]

    if (listeners) {
      this._listeners[event] = listeners.filter(
        /** @param {Listener} cb */
        cb => cb !== callback
      )
    }
  }
}
